!+
! Module: colsmooth_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task smoothes a column of a fits file.
! 
! Comments:
!	TODO:
!	- Use column units for vertical, row number??? for horizontal labels on plot?
!	- copy column units and comments to smoothed columns?
!
!****** should be i - j in doco (and equivalent in code)
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module colsmooth_mod

  use dal
  use param  
  use errorHandling
  use dataFilter
!  use constants

use temp_ssclib
!  use mathutils_mod
!  use plotutils_mod

use tests_mod

  implicit none 

! Command-line parameters:

  integer(int16) :: topHatWidth, numConvolverChannels
  integer(int16) :: maxSizeHoleToFill
  real(single) :: gaussianSigma
  character(255) :: inSetName, outSetName, convSetName, maskSetName, maskExpression, plotFileName
  character(20) :: inColName, outColName, inUncertsColName, outUncertsColName, convType, convColName, maskType, maskColName, pgDev
  logical :: withOutSet, recalculateUncerts, withPlotOutput, showUncert, normalisePlease
  real(single), pointer :: convolver(:)

  private 

  public colsmooth

contains 

  !---------------------------------------------------------------------
  subroutine colsmooth
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: values(:), uncerts(:)
    logical, pointer :: mask(:)
    logical, allocatable :: filledMask(:)
    integer :: status
    integer(int16) :: widthLargestHole, widthSmallestHole

    !-------------------------------------------------------------------
    ! Start

    call readParameters

!call test_make_spec_0
!return

    call getConvolver(convolver)

if (sum(convolver) .eq. 0.0) call error('zeroConvolverNorm'&
, 'The convolver elements sum to zero.')

!    widthLargestHole = size(convolver) - 1
    widthLargestHole = measureLargestHole(convolver)
    if (maxSizeHoleToFill .gt. widthLargestHole) then
      call warning('mustLimitInterpolation', 'The hole interpolation limit &
      &you have requested via the --fillholewidth parameter is too large &
      &for the convolver. It has been trimmed to the maximum possible size, &
      &ie 1 less than the span of non-zero values in the convolver.')

      maxSizeHoleToFill = widthLargestHole
    end if

    call getInSet(values, uncerts)
    call getMask(mask, int(size(values), int16))
    allocate(filledMask(size(mask)))
    call cyclicConvolution(values, convolver, mask, normalise=normalisePlease&
    , interpolationLimit=int(maxSizeHoleToFill), interpolationMask=filledMask&
    , uncerts=uncerts, status=status)

if (status .eq. -4) call error('zerosInConvolver'&
, 'Bad status return from convolution - probably because your convolver &
&has some zero or negative values.')

    call fitsOutput(values, uncerts, filledMask)
    if (withPlotOutput) call plotOutput(values, uncerts, filledMask)
    widthSmallestHole = measureSmallestHole(filledMask)

    deallocate(convolver, values, uncerts, mask, filledMask)

    write(errstr, '(a, i5, a)') 'The largest hole that can be filled &
    &using the present convolver is ', widthLargestHole, ' channels wide.'
    call message(AppMsg, Verbose)
    write(errstr, '(a, i5, a)') 'The smallest hole remaining is '&
    , widthSmallestHole, ' channels wide.'
    call message(AppMsg, Verbose)

!call addAngstroms !**** strictly a 1-off, for the estec poster!

  end subroutine colsmooth

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer(int16) :: i

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading parameters.')

    inSetName = stringParameter('intable')
    inColName = stringParameter('incolumn')
    withOutSet = booleanParameter('withouttable')
    if (withOutSet) then
      outSetName = stringParameter('outtable')
    end if
    outColName = stringParameter('outcolumn')
    recalculateUncerts = booleanParameter('witherrors')
    if (recalculateUncerts) then
      inUncertsColName = stringParameter('inerrorscolumn')
      outUncertsColName = stringParameter('outerrorscolumn')
    end if
    convType = stringParameter('convolvertype')
    select case(convType)
    case('set')
      convSetName = stringParameter('convolvertable')
      convColName = stringParameter('convolvercolumn')
    case('tophat')
      topHatWidth = intParameter('width')
    case('gaussian')
      gaussianSigma = realParameter('sigma')
    case('user')
      numConvolverChannels = parameterCount('convolver')
      allocate(convolver(numConvolverChannels))
      do i = 1, numConvolverChannels
        convolver(i) = realParameter('convolver', i - 1)
      end do
    case default
call error('badConvolverType', "Don't recognise the value of the --convolvertype parameter.")
    end select
    normalisePlease = booleanParameter('normalise')
    maxSizeHoleToFill = intParameter('fillholewidth')
    maskType = stringParameter('masktype')
    select case(maskType)
    case('set')
      maskSetName = stringParameter('masktable')
      maskColName = stringParameter('maskcolumn')
    case('expression')
      maskExpression = stringParameter('maskexpression')
    case('none')
! no action.
    case default
call error('badMaskType', "Don't recognise the value of the --masktype parameter.")
    end select
!    withPlotOutput = booleanParameter('withplot')
withPlotOutput = .false.
!!! Just for the time being, until I get time to test it properly.
    if (withPlotOutput) then
      showUncert = booleanParameter('showuncert')
      pgDev = stringParameter('pgdev')
      plotFileName = stringParameter('plotfile')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getConvolver(convolver)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolver(:)
    real(single), pointer :: dalConvolver(:)
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Making convolver.')

    select case(convType)
    case('set')
      set = dataSet(convSetName, READ)
      tab = table(set, convSetName)
      dalConvolver => real32Data(column(tab, convColName))
      allocate(convolver(size(dalConvolver)))
      convolver = dalConvolver
      call release(set)

    case('tophat')
      allocate(convolver(topHatWidth))
      convolver = 1.0

    case('gaussian')
      call makeGaussConvolver(convolver, gaussianSigma)

    case('user') ! then convolver vector is already loaded from the appropriate parameter.
if (numConvolverChannels .lt. 1) call error('convolverTooSmall', 'The convolver size is less than 1 channel!')
    case default
call error('badConvolverType', "Don't recognise the value of the --convolvertype parameter.")
    end select

  end subroutine getConvolver

  !---------------------------------------------------------------------
  subroutine makeGaussConvolver(convolver, sigma)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: convolver(:)
    real(single), intent(in) :: sigma

    integer(int16) :: convolverSize, halfConvolverSize, i
    real(single) :: x

    !-------------------------------------------------------------------
    ! Start

    convolverSize = floor(2.0 * 5.0 * sigma)
    halfConvolverSize = floor(convolverSize / 2.0)
    allocate(convolver(convolverSize))
    do i = 1, convolverSize
      x = real(i - 1 - halfConvolverSize)
      convolver(i) = exp(-x * x / 2.0 / sigma / sigma)
    end do

  end subroutine makeGaussConvolver

  !---------------------------------------------------------------------
  function measureLargestHole(convolver) result(largestHole)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: convolver(:)
    integer(int16) :: largestHole

    integer(int16) :: vsize, i, convStart, convFinis

    !-------------------------------------------------------------------
    ! Start

    vsize = size(convolver)
    if (vsize .le. 0) then
      largestHole = 0
  return
    end if

    convStart = vsize + 1 ! default
    do i = 1, vsize
      if (convolver(i) .ne. 0.0) then
        convStart = i
    exit
      end if
    end do
    convFinis = 0 ! default
    do i = vsize, 1, -1
      if (convolver(i) .ne. 0.0) then
        convFinis = i
    exit
      end if
    end do

    largestHole = convFinis - convStart
if (largestHole .lt. 0) call error('bug0', 'Should not happen! Contact the &
&code developer.')

  end function measureLargestHole

  !---------------------------------------------------------------------
  subroutine getInSet(values, uncerts)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: values(:), uncerts(:)
    real(double), pointer :: dalValues(:), dalUncerts(:)
    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: valuesCol, uncertsCol

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Reading inset.')

    set = dataSet(inSetName, READ)
    tab = table(set, inSetName)

    valuesCol = column(tab, inColName)
    call readColData(valuesCol, dalValues)
    allocate(values(size(dalValues)), uncerts(size(dalValues)))
    values = dalValues

    if (recalculateUncerts) then
      uncertsCol = column(tab, inUncertsColName)
      call readColData(uncertsCol, dalUncerts)
      uncerts = dalUncerts
    else
      uncerts = 0.0
    end if

    call release(set)

  end subroutine getInSet

  !---------------------------------------------------------------------
  subroutine getMask(mask, vsize)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: vsize
    logical, pointer :: mask(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    logical(BOOL), pointer :: boolMask(:)
    integer(int32) :: numOfRows, i
    integer(int32), dimension(:), pointer :: fromRow, toRow

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Making mask.')

    allocate(mask(vsize))

    select case(maskType)
    case('set')
      set = dataSet(maskSetName, READ)
      tab = table(set, maskSetName)
      boolMask => boolData(column(tab, maskColName))
      mask = boolMask
      call release(set)

    case('expression')
      set = dataSet(inSetName, READ)
      tab = table(set, inSetName)
      call initFilter(maskExpression)
      numOfRows = doSelect(tab, fromRow, toRow)
      mask = .false.
      do i = 0, size(fromRow) - 1
        mask(fromRow(i) + 1:toRow(i) + 1) = .true.
      end do

      call release(set)

    case('none')
      mask = .true.

    case default
call error('badMaskType', "Don't recognise the value of the --masktype parameter.")
    end select

  end subroutine getMask

  !---------------------------------------------------------------------
  subroutine fitsOutput(values, uncerts, mask)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: values(:), uncerts(:)
    logical, intent(in) :: mask(:)

    real(single), pointer :: dalValues(:), dalUncerts(:)
    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: valuesCol, uncertsCol
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64
    integer(int32) :: row

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Writing FITS output.')

    if (withOutSet) then
      set = dataSet(outSetName, CREATE)
      tab = addTable(set, outSetName, size(values))
      valuesCol = addColumn(tab, outColName, REAL32)
      dalValues => real32Data(valuesCol)
      dalValues = values
      do row = 0, size(values) - 1
        if(.not. mask(row + 1)) call setToNull(valuesCol, row)
      end do

      if (recalculateUncerts) then
        uncertsCol = addColumn(tab, outUncertsColName, REAL32)
        dalUncerts => real32Data(uncertsCol)
        dalUncerts = uncerts
        do row = 0, size(values) - 1
          if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
        end do
      end if

      call release(set)

    else ! modify inset:
      set = dataSet(inSetName, MODIFY)
      tab = table(set, inSetName)

      if (hasColumn(tab, outColName)) then
        valuesCol = column(tab, outColName)

        if (dataType(valuesCol) .eq. INTEGER8) then
          colDataInt8 => int8Data(valuesCol)
          colDataInt8 = convertToInt8(values)
          if (.not. nullDefined(valuesCol))&
            call setNullValue(valuesCol, int(huge(1_int8), int32))
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(valuesCol, row)
          end do

        else if (dataType(valuesCol) .eq. INTEGER16) then
          colDataInt16 => int16Data(valuesCol)
          colDataInt16 = convertToInt16(values)
          if (.not. nullDefined(valuesCol))&
            call setNullValue(valuesCol, int(huge(1_int16), int32))
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(valuesCol, row)
          end do

        else if (dataType(valuesCol) .eq. INTEGER32) then
          colDataInt32 => int32Data(valuesCol)
          colDataInt32 = convertToInt32(values)
          if (.not. nullDefined(valuesCol))&
            call setNullValue(valuesCol, huge(1_int32))
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(valuesCol, row)
          end do

        else if (dataType(valuesCol) .eq. REAL32) then
          colDataReal32 => real32Data(valuesCol)
          colDataReal32 = values
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(valuesCol, row)
          end do

        else ! this gives a DAL error if data type is none of the above.
          colDataReal64 => real64Data(valuesCol)
          colDataReal64 = real(values, double)
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(valuesCol, row)
          end do
        end if

      else ! inset doesn't have any such column.
        valuesCol = addColumn(tab, outColName, REAL32)
        dalValues => real32Data(valuesCol)
        dalValues = values
        do row = 0, size(values) - 1
          if(.not. mask(row + 1)) call setToNull(valuesCol, row)
        end do
      end if

      if (recalculateUncerts) then
        if (hasColumn(tab, inUncertsColName)) then
          uncertsCol = column(tab, inUncertsColName)

          if (dataType(uncertsCol) .eq. INTEGER8) then
            colDataInt8 => int8Data(uncertsCol)
            colDataInt8 = convertToInt8(uncerts)
            if (.not. nullDefined(uncertsCol))&
              call setNullValue(uncertsCol, int(huge(1_int8), int32))
            do row = 0, size(uncerts) - 1
              if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
            end do

          else if (dataType(uncertsCol) .eq. INTEGER16) then
            colDataInt16 => int16Data(uncertsCol)
            colDataInt16 = convertToInt16(uncerts)
            if (.not. nullDefined(uncertsCol))&
              call setNullValue(uncertsCol, int(huge(1_int16), int32))
            do row = 0, size(uncerts) - 1
              if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
            end do

          else if (dataType(uncertsCol) .eq. INTEGER32) then
            colDataInt32 => int32Data(uncertsCol)
            colDataInt32 = convertToInt32(uncerts)
            if (.not. nullDefined(uncertsCol))&
              call setNullValue(uncertsCol, huge(1_int32))
            do row = 0, size(uncerts) - 1
              if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
            end do

          else if (dataType(uncertsCol) .eq. REAL32) then
            colDataReal32 => real32Data(uncertsCol)
            colDataReal32 = uncerts
            do row = 0, size(uncerts) - 1
              if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
            end do

          else ! this gives a DAL error if data type is none of the above.
            colDataReal64 => real64Data(uncertsCol)
            colDataReal64 = real(uncerts, double)
            do row = 0, size(uncerts) - 1
              if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
            end do
          end if

        else ! inset doesn't have any such column.
          uncertsCol = addColumn(tab, outUncertsColName, REAL32)
          dalUncerts => real32Data(uncertsCol)
          dalUncerts = uncerts
          do row = 0, size(values) - 1
            if(.not. mask(row + 1)) call setToNull(uncertsCol, row)
          end do
        end if
      end if

      call release(set)
    end if

  end subroutine fitsOutput

  !---------------------------------------------------------------------
  subroutine plotOutput(values, uncerts, mask)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: values(:), uncerts(:)
    logical, intent(in) :: mask(:)

    integer :: status

    !-------------------------------------------------------------------
    ! Start

    if (recalculateUncerts) then
      call plotGraph(real(values), yIsNull=(.not. mask)&
      , device=trim(plotFileName)//trim(pgDev), yErrors=real(uncerts)&
      , label='Data set '//trim(inSetName)//', column '//trim(inColName)&
      , status=status)
!**** errors on status return?

    else
      call plotGraph(real(values), yIsNull=(.not. mask)&
      , device=trim(plotFileName)//trim(pgDev)&
      , label='Data set '//trim(inSetName)//', column '//trim(inColName)&
      , status=status)
!**** errors on status return?
    end if

  end subroutine plotOutput

  !---------------------------------------------------------------------
  function measureSmallestHole(mask) result(smallestHole)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) :: mask(:)
    integer(int16) :: smallestHole

    integer(int16) :: vsize, i, holeStart, holeSize
    logical :: inAhole

    !-------------------------------------------------------------------
    ! Start

    vsize = size(mask)
    if (vsize .le. 0) then
      smallestHole = 0
  return
    end if

    smallestHole = vsize ! default

    if (mask(1)) then
      inAhole = .false.

    else
      inAhole = .true.
      holeStart = 1
    end if

    do i = 2, vsize
      if (mask(i)) then
        if (inAhole) then
          holeSize = i - holeStart
          if (holeSize .lt. smallestHole) smallestHole = holeSize
          inAhole = .false.
        end if

      else if (.not. inAhole) then
        holeStart = i
        inAhole = .true.
      end if
    end do

    if (inAhole) then ! ie, mask(vsize) = .false.
      holeSize = vsize + 1 - holeStart
      if (holeSize .lt. smallestHole) smallestHole = holeSize

    else if (smallestHole .eq. vsize) then ! if the value in smallestHole were
    ! true, this combination (ie mask(vsize) .and. smallestHole .eq. vsize)
    ! would be impossible. Hence what this means is that no holes at all were
    ! found and that the true value of smallestHole is hence 0.
      smallestHole = 0
    end if

  end function measureSmallestHole

  !---------------------------------------------------------------------
  subroutine addAngstroms !**** strictly a 1-off, for the estec poster!
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: channelsCol, angstromsCol, kevCol
    real(double), pointer :: angstroms(:), kev(:)
    integer(int16), pointer :: channels(:)
    real(double) :: betaRef, betaWid, gratingDangstroms, alpha0,&
      srcDispArcMin, srcXdspArcMin, beta, alpha, pii
    integer(int32) :: row, order

    !-------------------------------------------------------------------
    ! Start

pii = 4.0 * atan2(1.0, 1.0)
order = -1

    set = dataSet(inSetName, READ)
    tab = table(set, inSetName)
    betaRef           = real64Attribute(tab, 'BETA_REF')
    betaWid           = real64Attribute(tab, 'BETA_WID')
    gratingDangstroms = real64Attribute(tab, 'LINE_SEP')
    alpha0            = real64Attribute(tab, 'ALPHANOM')
    srcDispArcMin     = real64Attribute(tab, 'RGS_DISP')
    srcXdspArcMin     = real64Attribute(tab, 'RGS_XDSP')
    call release(set)

    alpha = alpha0 + (srcDispArcMin / 60.0) * pii / 180.0

    if (withOutSet) then
      set = dataSet(outSetName, MODIFY)
      tab = table(set, outSetName)

    else ! modify inset:
      set = dataSet(inSetName, MODIFY)
      tab = table(set, inSetName)
    end if

    channelsCol = column(tab, 'CHANNEL')
    channels => int16Data(channelsCol)
    angstromsCol = addColumn(tab, 'ANGSTROM', REAL64)
    angstroms => real64Data(angstromsCol)
    kevCol = addColumn(tab, 'KEV', REAL64)
    kev => real64Data(kevCol)

    do row = 0, numberOfRows(tab) - 1
      beta = channels(row) * betaWid + betaRef
      angstroms(row) = order * gratingDangstroms * (cos(beta) - cos(alpha))
      kev(row) = angstroms2eV(angstroms(row)) / 1000.0
    end do

    call release(set)


  end subroutine addAngstroms

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module colsmooth_mod
