!+
! Module: temp_ssclib
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This contains subroutines it is intended to place eventually in ssclib mathutils_mod and plotutils_mod.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module temp_ssclib

  use dal
  use param  
  use errorHandling

  implicit none 

  private 

  real (kind=double), parameter :: SpeedOfLight = 2.99792458d8    ! m/sec
  real (kind=double), parameter :: PlanckH      = 6.62606876d-34  ! J sec
  real (kind=double), parameter :: ElectronQ    = 1.602176462d-19 ! coulombs

  public readColData
  public plotGraph
  public cyclicConvolution
  public convertToInt8
  public convertToInt16
  public convertToInt32

  public angstroms2eV

  interface plotGraph
    module procedure plotSingleGraph, plotManyGraphs
  end interface

  interface convertToInt8
    module procedure convertToInt8Array, convertToInt8Vector
  end interface

  interface convertToInt16
    module procedure convertToInt16Array, convertToInt16Vector
  end interface

  interface convertToInt32
    module procedure convertToInt32Array, convertToInt32Vector
  end interface

  character(3), parameter ::&
    defaultDevice = '/xs',&
    cmapModDevice = defaultDevice
  integer :: minClrIndex, maxClrIndex, numClrIndices
  real :: realNumCIminus1, realNumCI, oldXedges(1:256)!!!, newXedges(1:256)
  integer, parameter ::&
    numReservedClrs = 3,&
    DEV_NOT_VALID = -1,&
    NO_DEV_OPEN = 0,&
    DEV_NOT_OPEN = 1,&
    DEV_OPEN_NOT_SEL = 2,&
    DEV_OPEN_AND_SEL = 3
  real, parameter :: vpMinX = 0.07, vpMaxX = 0.999, vpMinY = 0.1, vpMaxY = 0.90

contains 

  !---------------------------------------------------------------------
  subroutine readColData(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 pointer colData.
!**** should be in ssclib!

    implicit none 

    type(ColumnT), intent(in) ::&
      col
    integer(kind=int8), dimension(:), pointer ::&
      colDataInt8
    integer(kind=int16), dimension(:), pointer ::&
      colDataInt16
    integer(kind=int32), dimension(:), pointer ::&
      colDataInt32
    real(kind=single), dimension(:), pointer ::&
      colDataReal32
    real(kind=double), dimension(:), pointer ::&
      colDataReal64,&
      colData

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

  end subroutine readColData

  !---------------------------------------------------------------------
  subroutine cyclicConvolution(vector, convolver, mask, padPlease&
  , normalise, interpolationLimit, interpolationMask, uncerts, status)
  !---------------------------------------------------------------------
!**** eventually put in ssclib/mathutils_mod

    implicit none

    real, intent(in)               :: convolver(:)
    real, intent(inout)            :: vector(:)
    real, intent(inout),  optional :: uncerts(:)
    logical, intent(in),  optional :: mask(:), padPlease, normalise
    integer, intent(in),  optional :: interpolationLimit
    integer, intent(out), optional :: status
    logical, intent(out), optional :: interpolationMask(:)

    integer :: vsize, convolverSize, i, ci, ri, localStatus, psize, halfConvolverSize,&
      interpolateHolesSmallerThan, holeStart, holeFinis
    real, allocatable :: paddedVector(:), output(:)
    real :: summ, weight, norm
    logical :: localPadPlease, localNormalise, inNullPatch
    logical, allocatable :: localMask(:), localInterpolationMask(:)

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default.

    vsize = size(vector)
    convolverSize = size(convolver)
    halfConvolverSize = floor(convolverSize / 2.0)

    if (present(padPlease)) then
      localPadPlease = padPlease
    else
      localPadPlease = .true.
    end if
    if (present(normalise)) then
      localNormalise = normalise
    else
      localNormalise = .true.
    end if
!print *, 'normalise:', localNormalise

    if (present(interpolationLimit)) then
      interpolateHolesSmallerThan = interpolationLimit
    else
      interpolateHolesSmallerThan = 0
    end if

    if (interpolateHolesSmallerThan .ge. convolverSize) then
      if (present(status)) status = -1
  return
    end if

    if (localPadPlease) then
      psize = vsize + convolverSize ! Pad doesn't need to be quite this big, but
      ! it won't hurt.
    else
      psize = vsize
    end if

    if (localNormalise) then
      norm = 1.0
    else
      norm = sum(convolver)
      if (norm .eq. 0.0) then
        if (present(status)) status = -2
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
    end if
!print *, 'norm=', norm

    allocate(paddedVector(psize), output(psize), localMask(psize)&
    , localInterpolationMask(psize))

    localMask = .false.
    if (present(mask)) then
      if (size(mask) .ne. vsize) then
        if (present(status)) status = -3
        deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
      end if
      localMask(1:vsize) = mask

    else
      localMask(1:vsize) = .true.
    end if

    ! Check all the holes to see if any are small enough to be interpolated:
    localInterpolationMask = localMask ! vector default
    if (.not. localMask(1)) then
      inNullPatch = .true.
      holeStart = 1
    else
      inNullPatch = .false.
    end if
    do i = 2, vsize
      if (inNullPatch .and. localMask(i)) then
        holeFinis = i - 1
        if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
          localInterpolationMask(holeStart:holeFinis) = .true.
        end if
        inNullPatch = .false.

      else if (.not. inNullPatch .and. .not. localMask(i)) then
        holeStart = i
        inNullPatch = .true.
      end if
    end do
    if (inNullPatch) then
      holeFinis = vsize
      if (holeFinis - holeStart + 1 .lt. interpolateHolesSmallerThan) then
        localInterpolationMask(holeStart:holeFinis) = .true.
      end if
    end if

    paddedVector(1:vsize) = vector
    output = 0.0 ! vector default.
    do i = 1, vsize
    if (.not. localInterpolationMask(i)) cycle
      summ = 0.0
      weight = 0.0
      do ci = 1, convolverSize
        ri = i - (ci - 1 - halfConvolverSize)
        if (ri .le. 0)     ri = ri + psize
        if (ri .gt. psize) ri = ri - psize
      if (.not. localMask(ri)) cycle
        summ = summ + convolver(ci) * paddedVector(ri)
        weight = weight + convolver(ci)
      end do
      if (weight .eq. 0.0)  then
        if (summ .eq. 0.0) then
          output(i) = 0.0

        else
          if (present(status)) status = -4
          deallocate(paddedVector, output, localMask, localInterpolationMask)
  return
        end if
      end if
      output(i) = summ * norm / weight
    end do

    vector = output(1:vsize)

    if (present(uncerts)) then
      paddedVector(1:vsize) = uncerts * uncerts
      output = 0.0 ! vector default.
      do i = 1, vsize
      if (.not. localInterpolationMask(i)) cycle
        summ = 0.0
        weight = 0.0
        do ci = 1, convolverSize
          ri = i - (ci - 1 - halfConvolverSize)
          if (ri .le. 0)     ri = ri + psize
          if (ri .gt. psize) ri = ri - psize
        if (.not. localMask(ri)) cycle
          summ = summ + convolver(ci) * convolver(ci) * paddedVector(ri)
          ! Recall that paddedVector contains the square of the uncertainties.

          weight = weight + convolver(ci)
        end do
        if (summ .eq. 0.0) then
          output(i) = 0.0

        else ! the circumstance in which summ .ne. 0 but weight .eq. 0 has
        ! already been dealt with above.
          output(i) = sqrt(summ) * norm / weight
        end if
      end do

      uncerts = output(1:vsize)
    end if

    if (present(interpolationMask))&
      interpolationMask = localInterpolationMask(1:vsize)

    deallocate(paddedVector, output, localMask, localInterpolationMask)

    if (present(status)) status = localStatus

  end subroutine cyclicConvolution

  !---------------------------------------------------------------------
  function convertToInt8Array(real32array) result(int8array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int8) :: int8array(size(real32array, 1), size(real32array, 2))
    integer(int8) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int8))
    min = 1 - max
    where(real32array .ge. max)
      int8array = max
    elsewhere(real32array .le. min)
      int8array = min
    elsewhere
      int8array = int(real32array, int8)
    end where

  end function convertToInt8Array

  !---------------------------------------------------------------------
  function convertToInt16Array(real32array) result(int16array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int16) :: int16array(size(real32array, 1), size(real32array, 2))
    integer(int16) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int16))
    min = 1 - max
    where(real32array .ge. max)
      int16array = max
    elsewhere(real32array .le. min)
      int16array = min
    elsewhere
      int16array = int(real32array, int16)
    end where

  end function convertToInt16Array

  !---------------------------------------------------------------------
  function convertToInt32Array(real32array) result(int32array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32array(:,:)
    integer(int32) :: int32array(size(real32array, 1), size(real32array, 2))
    integer(int32) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int32))
    min = 1 - max
    where(real32array .ge. max)
      int32array = max
    elsewhere(real32array .le. min)
      int32array = min
    elsewhere
      int32array = int(real32array, int32)
    end where

  end function convertToInt32Array

  !---------------------------------------------------------------------
  function convertToInt8Vector(real32Vector) result(int8Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int8) :: int8Vector(size(real32Vector))
    integer(int8) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int8))
    min = 1 - max
    where(real32Vector .ge. max)
      int8Vector = max
    elsewhere(real32Vector .le. min)
      int8Vector = min
    elsewhere
      int8Vector = int(real32Vector, int8)
    end where

  end function convertToInt8Vector

  !---------------------------------------------------------------------
  function convertToInt16Vector(real32Vector) result(int16Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int16) :: int16Vector(size(real32Vector))
    integer(int16) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int16))
    min = 1 - max
    where(real32Vector .ge. max)
      int16Vector = max
    elsewhere(real32Vector .le. min)
      int16Vector = min
    elsewhere
      int16Vector = int(real32Vector, int16)
    end where

  end function convertToInt16Vector

  !---------------------------------------------------------------------
  function convertToInt32Vector(real32Vector) result(int32Vector)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: real32Vector(:)
    integer(int32) :: int32Vector(size(real32Vector))
    integer(int32) :: min, max

    !-------------------------------------------------------------------
    ! Start

    max = huge(int(1, int32))
    min = 1 - max
    where(real32Vector .ge. max)
      int32Vector = max
    elsewhere(real32Vector .le. min)
      int32Vector = min
    elsewhere
      int32Vector = int(real32Vector, int32)
    end where

  end function convertToInt32Vector

  !---------------------------------------------------------------------
  subroutine plotSingleGraph(y, x, yErrors, yIsNull, minX, maxX, minY, maxY&
  , suppressYzero, device, rainbow, style, xLabel, yLabel, label, wait&
  , deviceId, status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)              :: y(:)
    real,         intent(in),    optional :: x(:),&
                                             yErrors(:),&
                                             minX, maxX, minY, maxY
    character(*), intent(in),    optional :: device,&
                                             style,&
                                             xLabel,&
                                             yLabel,&
                                             label
    logical,      intent(in),    optional :: yIsNull(:),&
                                             suppressYzero,&
                                             rainbow,&
                                             wait
    integer,      intent(inout), optional :: deviceId
    integer,      intent(out),   optional :: status

    integer :: i, vectorSize, localDevId, localStatus
    real, dimension(:), allocatable :: localX, localYerrors
    logical, dimension(:), allocatable :: localYisNull
    character(255) :: localDevice, localStyle, localXlabel, localYlabel,&
      localLabel
    logical :: localSuppressYzero, withYerrors, localRainbow, localWait,&
      withMinX, withMaxX, withMinY, withMaxY
    real :: localY(1:size(y), 1), localMinX, localMaxX, localMinY, localMaxY

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    vectorSize = size(y)
    localY(1:vectorSize, 1) = y

    if (present(x)) then
      allocate(localX(size(x)))
      localX = x
    else
      allocate(localX(vectorSize))
      do i = 1, vectorSize
        localX(i) = real(i)
      end do
    end if

    if (present(yErrors)) then
      allocate(localYerrors(size(yErrors)))
      localYerrors = yErrors
      withYerrors = .true.
    else
      allocate(localYerrors(vectorSize))
      localYerrors = 0.0
      withYerrors = .false.
    end if

    if (present(minX)) then
      localMinX = minX
      withMinX = .true.
    else
      withMinX = .false.
    end if
    if (present(maxX)) then
      localMaxX = maxX
      withMaxX = .true.
    else
      withMaxX = .false.
    end if
    if (present(minY)) then
      localMinY = minY
      withMinY = .true.
    else
      withMinY = .false.
    end if
    if (present(maxY)) then
      localMaxY = maxY
      withMaxY = .true.
    else
      withMaxY = .false.
    end if

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = trim(style)
    else
      localStyle = 'line'
    end if

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    if (present(yIsNull)) then
      allocate(localYisNull(size(yIsNull)))
      localYisNull = yIsNull
    else
      allocate(localYisNull(vectorSize))
      localYisNull = .false.
    end if

    if (present(suppressYzero)) then
      localSuppressYzero = suppressYzero
    else
      localSuppressYzero = .false.
    end if

    if (present(rainbow)) then ! suppresses compiler warnings.
    end if
    localRainbow = .false.

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(deviceId)) then
      localDevId = deviceId
    else
      localDevId = 0 ! not a valid number - this signals to kernel that a
      ! new device is required.
    end if

    localStatus = plotGraphKernel(localY, localX, withYerrors, localYerrors&
    , localYisNull, withMinX, localMinX, withMaxX, localMaxX, withMinY&
    , localMinY, withMaxY, localMaxY, localSuppressYzero, localRainbow&
    , localDevice, localStyle&
    , localXlabel, localYlabel, localLabel, localWait, localDevId)

    if (present(deviceId)) deviceId = localDevId
    if (present(status)) status = localStatus

  end subroutine plotSingleGraph

  !---------------------------------------------------------------------
  subroutine plotManyGraphs(y, x, yErrors, yIsNull, minX, maxX, minY, maxY&
  , suppressYzero, device, rainbow, style, xLabel, yLabel, label, wait&
  , deviceId, status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)              :: y(:,:)
    real,         intent(in),    optional :: x(:),&
                                             yErrors(:),&
                                             minX, maxX, minY, maxY

    character(*), intent(in),    optional :: device,&
                                             style,&
                                             xLabel,&
                                             yLabel,&
                                             label
    logical,      intent(in),    optional :: yIsNull(:),&
                                             suppressYzero,&
                                             rainbow,&
                                             wait
    integer,      intent(inout), optional :: deviceId
    integer,      intent(out),   optional :: status

    integer :: i, vectorSize, localDevId, localStatus
    real, dimension(:), allocatable :: localX, localYerrors
    logical, dimension(:), allocatable :: localYisNull
    character(255) :: localDevice, localStyle, localXlabel, localYlabel,&
      localLabel
    logical :: localSuppressYzero, withYerrors, localRainbow,&
      localWait,&
      withMinX, withMaxX, withMinY, withMaxY
    real :: localMinX, localMaxX, localMinY, localMaxY

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0 ! default

    vectorSize = size(y, 1)

    if (present(x)) then
      allocate(localX(size(x)))
      localX = x
    else
      allocate(localX(vectorSize))
      do i = 1, vectorSize
        localX(i) = real(i)
      end do
    end if

    if (present(yErrors)) then ! suppresses compiler warnings.
    end if
    allocate(localYerrors(vectorSize))
    localYerrors = 0.0
    withYerrors = .false.

    if (present(minX)) then
      localMinX = minX
      withMinX = .true.
    else
      withMinX = .false.
    end if
    if (present(maxX)) then
      localMaxX = maxX
      withMaxX = .true.
    else
      withMaxX = .false.
    end if
    if (present(minY)) then
      localMinY = minY
      withMinY = .true.
    else
      withMinY = .false.
    end if
    if (present(maxY)) then
      localMaxY = maxY
      withMaxY = .true.
    else
      withMaxY = .false.
    end if

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then ! suppresses compiler warnings.
    end if
    localStyle = 'line'

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    if (present(yIsNull)) then ! suppresses compiler warnings.
    end if
    allocate(localYisNull(vectorSize))
    localYisNull = .false.

    if (present(suppressYzero)) then
      localSuppressYzero = suppressYzero
    else
      localSuppressYzero = .false.
    end if

    if (present(rainbow)) then
      localRainbow = rainbow
    else
      localRainbow = .false.
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(deviceId)) then
      localDevId = deviceId
    else
      localDevId = 0 ! not a valid number - this signals to kernel that a
      ! new device is required.
    end if

    localStatus = plotGraphKernel(y, localX, withYerrors, localYerrors&
    , localYisNull, withMinX, localMinX, withMaxX, localMaxX, withMinY&
    , localMinY, withMaxY, localMaxY, localSuppressYzero, localRainbow&
    , localDevice, localStyle&
    , localXlabel, localYlabel, localLabel, localWait, localDevId)

    if (present(deviceId)) deviceId = localDevId
    if (present(status)) status = localStatus

  end subroutine plotManyGraphs

  !---------------------------------------------------------------------
  function plotGraphKernel(y, x, withYerrors, yErrors, yIsNull&
  , withMinX, minX, withMaxX, maxX, withMinY, minY, withMaxY, maxY&
  , suppressYzero, multiColoured, device, style, xLabel, yLabel, label&
  , wait, deviceId) result(status)
  !---------------------------------------------------------------------
!**** there is a potential problem with the device ids which could arise as follows: say the user has already opened a pg device, deviceId=a. He wants to open another, leave it open for further plots. So he calls plotGraph with the deviceId argument present (because its output value is required) but not initialised. If by chance the value happens to be 'a', then the previous plot will be overdrawn rather than a new one created. This can be avoided by the user initialising deviceId to zero or less. This should be documented.

    implicit none

    real, intent(in) :: y(:,:)
    real, intent(inout) :: x(:), yErrors(:), minX, maxX, minY, maxY
! These are 'inout' just so I don't have to have separate names.
    character(*), intent(in) :: device, style, xLabel, yLabel, label
    logical, intent(inout) :: yIsNull(:)
! This is 'inout' just so I don't have to have a separate name.
    logical, intent(in) :: withYerrors, suppressYzero,&
      multiColoured, wait, withMinX, withMaxX, withMinY, withMaxY
    integer, intent(inout) :: deviceId
    integer :: status

    integer :: pgopen, i, firstNonNull, iStart, iFinis,&
      vectorSize, numPlots, plotNum, j, deviceStatus, lastX,&
      foreignStatus, minFS, maxFS, statusMin, statusMax,&
      plotCIs(size(y, 2)), nullCI, axesCI, previousDevId
    logical :: allYnulls, noYnulls, inNullPatch, openNewDevice
    real, allocatable :: localY(:,:)
    real :: xEdgeWidth, yEdgeWidth

logical :: rebin !*** make it an argument when it is shown to work ok.

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    statusMin = -6
    statusMax = 8

    deviceStatus = queryDeviceStatus(deviceId)
    vectorSize = size(y, 1)
    numPlots   = size(y, 2)

!.......................................................................
! Checks:

    if (deviceStatus .eq. DEV_NOT_VALID) then
      status = -1
  return
    end if

    if (size(x) .lt. vectorSize) then
      status = -2
  return
    else if (size(x) .gt. vectorSize) then
      status = status + 1
    end if
    if (size(yErrors) .lt. vectorSize) then
      status = -3
  return
    else if (size(yErrors) .gt. vectorSize) then
      status = status + 2
    end if
    if (size(yIsNull) .lt. vectorSize) then
      status = -4
  return
    else if (size(yIsNull) .gt. vectorSize) then
      status = status + 4
    end if

    allYnulls = .true.
    do i = 1, vectorSize
      if (.not. yIsNull(i)) then
        allYnulls = .false.
        firstNonNull = i
    exit
      end if
    end do
    if (allYnulls) then
      status = status + 8
    end if

    noYnulls = .true.
    do i = 1, vectorSize
      if (yIsNull(i)) then
        noYnulls = .false.
    exit
      end if
    end do
    if (.not. noYnulls .and. (numPlots .gt. 1 .or. style .eq. 'polygon')) then
      status = -5
    end if

!.......................................................................
! Get extrema:

    if (.not. withMinX) minX = minval(x)
    if (.not. withMaxX) maxX = maxval(x)

    if (allYnulls) then
      minY = 0.0
      maxY = 1.0

    else
      if (numPlots .gt. 1) then
        if (.not. withMinY) minY = minval(y)
        if (.not. withMaxY) maxY = maxval(y)

      else if (withYerrors) then
        if (.not. withMinY) then
          minY = y(firstNonNull, 1) - yErrors(firstNonNull) 
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              minY = min(minY, y(i, 1) - yErrors(i))
            end if
          end do
        end if
        if (.not. withMaxY) then
          maxY = y(firstNonNull, 1) - yErrors(firstNonNull) 
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              maxY = max(maxY, y(i, 1) + yErrors(i))
            end if
          end do
        end if

      else
        if (.not. withMinY) then
          minY = y(firstNonNull, 1)
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              minY = min(minY, y(i, 1))
            end if
          end do
        end if
        if (.not. withMaxY) then
          maxY = y(firstNonNull, 1)
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              maxY = max(maxY, y(i, 1))
            end if
          end do
        end if
      end if
      if (.not. suppressYzero) then
        if (.not. withMinY) minY = min(0.0, minY)
        if (.not. withMaxY) maxY = max(0.0, maxY)
      end if
    end if ! allYnulls

!.......................................................................
! Open or select device:

    if (deviceStatus .eq. DEV_NOT_OPEN&
    .or. deviceStatus .eq. DEV_OPEN_NOT_SEL) then
      call pgqid(previousDevId)
    end if

    openNewDevice = .false. ! default
    if (deviceStatus .eq. NO_DEV_OPEN .or. deviceStatus .eq. DEV_NOT_OPEN) then
      openNewDevice = .true.
      deviceId = pgopen(trim(device))
      call pgsvp(vpMinX, vpMaxX, vpMinY, vpMaxY)
      call pgscr(2, 0.5, 0.5, 0.5)

    else if (deviceStatus .eq. DEV_OPEN_NOT_SEL) then
      call pgslct(deviceId)
    end if

!.......................................................................
! Set up fill style, colour:

    call pgsfs(2) ! unfilled.

    call setupColours(multiColoured, plotCIs, nullCI, axesCI, foreignStatus&
    , minFS, maxFS)
    call combineStatuses(foreignStatus, minFS, maxFS&
    , status, statusMin, statusMax)
    if (status .lt. 0) then
      call pgclos
  return
    end if

!.......................................................................
! Draw axes, box and labels:

    xEdgeWidth = 0.0
    yEdgeWidth = 0.05 * (maxY - minY)

    if (openNewDevice) then
      call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
      , minY - yEdgeWidth, maxY + yEdgeWidth)

      call pgsci(axesCI)
      call pgmove(0.0, minY - yEdgeWidth)
      call pgdraw(0.0, maxY + yEdgeWidth)
      call pgmove(minX - xEdgeWidth, 0.0)
      call pgdraw(maxX + xEdgeWidth, 0.0)
      call pgsci(1)
      call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
      call pglab(trim(xLabel), trim(yLabel), trim(label))
    end if

!.......................................................................
! Fork into different plot styles:

    select case(style)
    case('line', 'lines', 'linear')
rebin = .true.
      allocate(localY(vectorSize, numPlots))
!call test_makeXY(x, localY, yIsNull, yErrors)

      localY = y
      if (rebin) then
        call rebinX(x, localY, yIsNull, yErrors, lastX)
      else
        lastX = vectorSize
      end if

      if (numPlots .gt. 1) then
        do plotNum = 1, numPlots
          call pgsci(plotCIs(plotNum))
          call pgline(lastX, x(1:lastX), y(1:lastX, plotNum))
        end do

      else
        if (yIsNull(1)) then
          inNullPatch = .true.
        else
          inNullPatch = .false.
        end if
        iStart = 1
        do i = 2, lastX
          if (inNullPatch .and. .not. yIsNull(i)) then
            iFinis = i
            call pgsci(nullCI)
            call pgmove(x(iStart), 0.0)
            call pgdraw(x(iFinis), 0.0)
            iStart = i
            inNullPatch = .false.

          else if (.not. inNullPatch .and. yIsNull(i)) then
            iFinis = i - 1
            call pgsci(plotCIs(1))
            if (withYerrors) then
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
              do j = iStart, iFinis
                call pgmove(x(j), y(j, 1) - yErrors(j))
                call pgdraw(x(j), y(j, 1) + yErrors(j))
              end do

            else
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
            end if
            iStart = i - 1
            inNullPatch = .true.
          end if
        end do

        if (inNullPatch) then
          iFinis = lastX
          call pgsci(nullCI)
          call pgmove(x(iStart), 0.0)
          call pgdraw(x(iFinis), 0.0)

        else if (.not. inNullPatch) then
          iFinis = lastX
          call pgsci(plotCIs(1))
          if (withYerrors) then
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
            do j = iStart, iFinis
              call pgmove(x(j), y(j, 1) - yErrors(j))
              call pgdraw(x(j), y(j, 1) + yErrors(j))
            end do

          else
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
          end if
        end if
      end if

      deallocate(localY)

    case('point', 'points')
!******* add square, diamond, circle??

      if (withYerrors) then
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgmove(x(i), y(i, 1) - yErrors(i))
            call pgdraw(x(i), y(i, 1) + yErrors(i))
          end if
        end do

      else
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgpt1(x(i), y(i, 1), -1)
          end if
        end do
      end if

    case('polygon')
      do plotNum = 1, numPlots
        call pgsci(plotCIs(plotNum))
        call pgpoly(vectorSize, x(1:vectorSize), y(1:vectorSize, plotNum))
      end do

    case default
      status = -6
      call pgclos
  return
    end select

!*** figure out some more what we want to happen here:
    if (wait .and. .not. openNewDevice) then
      call pgslct(previousDevId)
    else if (.not. wait) then
      call pgclos
    end if

  end function plotGraphKernel

!  !---------------------------------------------------------------------
!  subroutine test_makeXY(x, y, yIsNull, yErrors)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    real, intent(inout) :: x(:), y(:,:), yErrors(:)
!    logical, intent(inout) :: yIsNull(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end subroutine test_makeXY

  !---------------------------------------------------------------------
  subroutine setupColours(multiColoured, plotCIs, nullCI, axesCI, status, statusMin, statusMax)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) :: multiColoured
    integer, intent(out) :: plotCIs(:), nullCI, axesCI, status
    integer, intent(out), optional :: statusMin, statusMax

    integer :: foreignStatus, minFS, maxFS, numPlots, plotNum
    real :: frac, redVal, grnVal, bluVal

    !-------------------------------------------------------------------
    ! Start

    status = 0

    numPlots = size(plotCIs)

    call findClrLimits(foreignStatus, minFS, maxFS) ! sets global integers
    ! minClrIndex, maxClrIndex, numClrIndices.

    call combineStatuses(foreignStatus, minFS, maxFS&
    , status, statusMin, statusMax)
  if (status .lt. 0) return

    if (numPlots + 5 .gt. numClrIndices) then
    ! Extra colours needed for (i) nulls, (ii) frame (fg), (iii) bg, (iv) extra plot, (v) axes.
      status = status + 1

      plotCIs = maxClrIndex ! vector op
      nullCI = maxClrIndex

    else
      nullCI = 3
      call pgscr(nullCI, 0.3, 1.0, 0.75)
      if (numPlots .eq. 1 .or. .not. multiColoured) then
!        plotCIs = 4
!        call pgscr(4, 1.0, 0.95, 0.7)
!!***** check that first colour is the same whether many plots or one.

plotCIs = 1
nullCI = 1
axesCI = 1

      else ! set up rainbow colours:
        do plotNum = 1, numPlots
          plotCIs(plotNum) = plotNum + 4
          frac = (plotNum - 1) / real(numPlots - 1)
          if (frac .lt. 0.25) then
            redVal = 1.0
            grnVal = frac / 0.25
            bluVal = 0.0
          else if (frac .lt. 0.5) then
            redVal = 1.0 - (frac - 0.25) / 0.25
            grnVal = 1.0
            bluVal = 0.0
          else if (frac .lt. 0.75) then
            redVal = 0.0
            grnVal = 1.0 - (frac - 0.5) / 0.25
            bluVal = (frac - 0.5) / 0.25
          else
            redVal = 0.7 * (frac - 0.75) / 0.25
            grnVal = 0.0
            bluVal = 1.0 - 0.3 * (frac - 0.75) / 0.25
          end if
          call pgscr(plotCIs(plotNum), redVal, grnVal, bluVal)
        end do
      end if
    end if

  end subroutine setupColours

  !---------------------------------------------------------------------
  subroutine combineStatuses(foreignStatus, foreignStatusMin&
  , foreignStatusMax, status, statusMin, statusMax)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: foreignStatus, foreignStatusMin, foreignStatusMax
    integer, intent(inout) :: status, statusMin, statusMax

    !-------------------------------------------------------------------
    ! Start

    if (foreignStatus .lt. 0) then ! foreignStatus dominates.
      status = statusMin + foreignStatus

    else if (foreignStatus .gt. 0) then ! status < 0 dominates.
      if (status .ge. 0) then
        status = status + 2 * statusMax * foreignStatus
      end if
    end if

    statusMin = statusMin + foreignStatusMin
    statusMax = 2 * statusMax * foreignStatusMax

  end subroutine combineStatuses

  !---------------------------------------------------------------------
  subroutine rebinX(x, y, yIsNull, yErrors, newI)
  !---------------------------------------------------------------------
  ! It is assumed that x values are in increasing order and all within the window X dimensions. The last x value must be _less_ than the max window X (wldXmax).

    implicit none

    real, intent(inout) :: x(:), y(:,:), yErrors(:)
    logical, intent(inout) :: yIsNull(:)
    integer, intent(out) :: newI

    integer :: vectorSize, numPlots, vpXmin, vpXmax, vpYmin, vpYmax,&
      wldXmin, wldXmax, wldYmin, wldYmax, oldI, bin, plotNum,&
      firstInBin, lastInBin, numNullsInBin, totalNumInBin
    real :: realDeltaVpX, botFrac, binBot, topFrac, binTop
    logical :: allDone

    !-------------------------------------------------------------------
    ! Start

    vectorSize = size(y, 1)
    numPlots   = size(y, 2)

    call pgqvsz(3, vpXmin, vpXmax, vpYmin, vpYmax)
    call pgqwin(wldXmin, wldXmax, wldYmin, wldYmax)
    realDeltaVpX = real(1 + vpXmax - vpXmin)

    allDone = .false.
    newI = 0
    oldI = 1
  if (oldI .gt. vectorSize) return

    do bin = vpXmin, vpXmax
      botFrac = (bin - vpXmin) / realDeltaVpX
      binBot = wldXmin * (1.0 - botFrac) + wldXmax * botFrac
      topFrac = (1 + bin - vpXmin) / realDeltaVpX
      binTop = wldXmin * (1.0 - topFrac) + wldXmax * topFrac

      firstInBin = oldI
      numNullsInBin = 0
      do while(x(oldI) .lt. binTop)
        totalNumInBin = totalNumInBin + 1
        if (yIsNull(oldI)) numNullsInBin = numNullsInBin + 1
        oldI = oldI + 1
        if (oldI .gt. vectorSize) then
          allDone = .true.
      exit
        end if
      end do
      lastInBin = oldI

      totalNumInBin = lastInBin - firstInBin
      if (totalNumInBin .gt. 0) then
        newI = newI + 1
        if (numNullsInBin .eq. totalNumInBin) then
          yIsNull(newI) = .true.
          ! forget about the other y values.

        else
          do plotNum = 1, numPlots
            where(yIsNull(firstInBin:lastInBin))
              y(firstInBin:lastInBin, plotNum) = 0.0
              yErrors(firstInBin:lastInBin) = 0.0
            end where
          end do
          y(newI,:) = sum(y(firstInBin:lastInBin,:))&
          / real(totalNumInBin - numNullsInBin)

          yErrors(newI) = sqrt(sum(yErrors(firstInBin:lastInBin)&
          * yErrors(firstInBin:lastInBin))&
          / real(totalNumInBin - numNullsInBin))

          yIsNull(newI) = .false.
        end if

        x(newI) = (binBot + binTop) / 2.0
      end if

    if (allDone) exit
    end do

  end subroutine rebinX

  !---------------------------------------------------------------------
  function plotGraphKernel_old(y, x, withYerrors, yErrors, yIsNull&
  , withMinX, minX, withMaxX, maxX, withMinY, minY, withMaxY, maxY&
  , suppressYzero, multiColoured, device, style, xLabel, yLabel, label&
  , wait, deviceId) result(status)
  !---------------------------------------------------------------------
!**** there is a potential problem with the device ids which could arise as follows: say the user has already opened a pg device, deviceId=a. He wants to open another, leave it open for further plots. So he calls plotGraph with the deviceId argument present (because its output value is required) but not initialised. If by chance the value happens to be 'a', then the previous plot will be overdrawn rather than a new one created. This can be avoided by the user initialising deviceId to zero or less. This should be documented.

    implicit none

    real, intent(in) :: x(:), y(:,:), yErrors(:)
    real, intent(inout) :: minX, maxX, minY, maxY
! These are 'inout' just so I don't have to have separate names.
    character(*), intent(in) :: device, style, xLabel, yLabel, label
    logical, intent(in) :: withYerrors, yIsNull(:), suppressYzero,&
      multiColoured, wait, withMinX, withMaxX, withMinY, withMaxY
    integer, intent(inout) :: deviceId
    integer :: status

    integer :: pgopen, i, firstNonNull, iStart, iFinis,&
      vectorSize, numPlots, plotNum, j, deviceStatus,&
      plotCIs(size(y, 2)), nullCI, previousDevId
!!!, answerLength, slashPosition
    real :: xEdgeWidth, yEdgeWidth,&
      frac, redVal, grnVal, bluVal
    logical :: allYnulls, noYnulls, inNullPatch, openNewDevice
!    character(3) :: answer
!    character(len(trim(device))) :: trueDevice

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    deviceStatus = queryDeviceStatus(deviceId)
    openNewDevice = .false.
    if (deviceStatus .le. 1) then
      openNewDevice = .true.
    else if (deviceStatus .eq. 2) then
      call pgslct(deviceId)
    end if

    vectorSize = size(y, 1)
    numPlots   = size(y, 2)

    if (size(x) .lt. vectorSize) then
      status = -1
  return
    else if (size(x) .gt. vectorSize) then
      status = status + 1
    end if
    if (size(yErrors) .lt. vectorSize) then
      status = -2
  return
    else if (size(yErrors) .gt. vectorSize) then
      status = status + 2
    end if
    if (size(yIsNull) .lt. vectorSize) then
      status = -3
  return
    else if (size(yIsNull) .gt. vectorSize) then
      status = status + 4
    end if

    allYnulls = .true.
    do i = 1, vectorSize
      if (.not. yIsNull(i)) then
        allYnulls = .false.
        firstNonNull = i
    exit
      end if
    end do
    if (allYnulls) then
      status = status + 8
    end if

    noYnulls = .true.
    do i = 1, vectorSize
      if (yIsNull(i)) then
        noYnulls = .false.
    exit
      end if
    end do
    if (.not. noYnulls .and. (numPlots .gt. 1 .or. style .eq. 'polygon')) then
      status = -4
    end if

    if (.not. withMinX) minX = minval(x)
    if (.not. withMaxX) maxX = maxval(x)

    if (allYnulls) then
      minY = 0.0
      maxY = 1.0

    else
      if (numPlots .gt. 1) then
        if (.not. withMinY) minY = minval(y)
        if (.not. withMaxY) maxY = maxval(y)

      else if (withYerrors) then
        if (.not. withMinY) then
          minY = y(firstNonNull, 1) - yErrors(firstNonNull) 
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              minY = min(minY, y(i, 1) - yErrors(i))
            end if
          end do
        end if
        if (.not. withMaxY) then
          maxY = y(firstNonNull, 1) - yErrors(firstNonNull) 
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              maxY = max(maxY, y(i, 1) + yErrors(i))
            end if
          end do
        end if

      else
        if (.not. withMinY) then
          minY = y(firstNonNull, 1)
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              minY = min(minY, y(i, 1))
            end if
          end do
        end if
        if (.not. withMaxY) then
          maxY = y(firstNonNull, 1)
          do i = firstNonNull + 1, vectorSize
            if (.not. yIsNull(i)) then
              maxY = max(maxY, y(i, 1))
            end if
          end do
        end if
      end if
      if (.not. suppressYzero) then
        if (.not. withMinY) minY = min(0.0, minY)
        if (.not. withMaxY) maxY = max(0.0, maxY)
      end if
    end if ! allYnulls

    if (openNewDevice) then
      deviceId = pgopen(trim(device))
      call pgsvp(0.07, 0.999, 0.1, 0.90)
!***** use the standard global values.
      call pgscr(2, 0.5, 0.5, 0.5)
    end if
    call pgsfs(2) ! unfilled.

    call findClrLimits(status)
  if (status .lt. 0) return
!*** make status numbers consistent.
!print *, minClrIndex, maxClrIndex, numClrIndices
    if (numPlots + 5 .gt. numClrIndices) then
    ! Extra colours needed for (i) nulls, (ii) frame (fg), (iii) bg, (iv) extra plot, (v) axes.
print *, 'max ci = ', maxClrIndex
!**** set a warning status value
      plotCIs = maxClrIndex ! vector op
      nullCI = maxClrIndex
    else
      nullCI = 3
      call pgscr(nullCI, 0.3, 1.0, 0.75)
      if (numPlots .eq. 1 .or. .not. multiColoured) then
!        plotCIs = 4
!        call pgscr(4, 1.0, 0.95, 0.7)
!!***** check that first colour is the same whether many plots or one.

plotCIs = 1
nullCI = 1

      else
        do plotNum = 1, numPlots
          plotCIs(plotNum) = plotNum + 4
          frac = (plotNum - 1) / real(numPlots - 1)
!          if (frac .lt. 0.4) then
          if (frac .lt. 0.25) then
            redVal = 1.0
!            grnVal = frac / 0.4
            grnVal = frac / 0.25
            bluVal = 0.0
!          else if (frac .lt. 0.6) then
          else if (frac .lt. 0.5) then
!            redVal = 1.0 - (frac - 0.4) / 0.2
            redVal = 1.0 - (frac - 0.25) / 0.25
            grnVal = 1.0
            bluVal = 0.0
!          else if (frac .lt. 0.8) then
          else if (frac .lt. 0.75) then
            redVal = 0.0
!            grnVal = 1.0 - (frac - 0.6) / 0.2
!            bluVal = (frac - 0.6) / 0.2
            grnVal = 1.0 - (frac - 0.5) / 0.25
            bluVal = (frac - 0.5) / 0.25
          else
!            redVal = 0.7 * (frac - 0.8) / 0.2
            redVal = 0.7 * (frac - 0.75) / 0.25
            grnVal = 0.0
!            bluVal = 1.0 - 0.3 * (frac - 0.8) / 0.2
            bluVal = 1.0 - 0.3 * (frac - 0.75) / 0.25
          end if
          call pgscr(plotCIs(plotNum), redVal, grnVal, bluVal)
!print *, plotNum, plotCIs(plotNum), redVal, grnVal, bluVal
        end do
      end if
    end if

    select case(style)
    case('line', 'lines', 'linear')
      xEdgeWidth = 0.0
      yEdgeWidth = 0.05 * (maxY - minY)

      if (openNewDevice) then
        call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
        , minY - yEdgeWidth, maxY + yEdgeWidth)

        call pgsci(2)
        call pgmove(0.0, minY - yEdgeWidth)
        call pgdraw(0.0, maxY + yEdgeWidth)
        call pgmove(minX - xEdgeWidth, 0.0)
        call pgdraw(maxX + xEdgeWidth, 0.0)
        call pgsci(1)
        call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
        call pglab(trim(xLabel), trim(yLabel), trim(label))
      end if

      if (numPlots .gt. 1) then
        do plotNum = 1, numPlots
!print *, plotNum
          call pgsci(plotCIs(plotNum))
          call pgline(vectorSize, x(1:vectorSize), y(1:vectorSize, plotNum))
        end do

      else
        if (yIsNull(1)) then
          inNullPatch = .true.
        else
          inNullPatch = .false.
        end if
        iStart = 1
        do i = 2, vectorSize
          if (inNullPatch .and. .not. yIsNull(i)) then
            iFinis = i
            call pgsci(nullCI)
            call pgmove(x(iStart), 0.0)
            call pgdraw(x(iFinis), 0.0)
            iStart = i
            inNullPatch = .false.

          else if (.not. inNullPatch .and. yIsNull(i)) then
            iFinis = i - 1
            call pgsci(plotCIs(1))
            if (withYerrors) then
!***** if too many x points, bin up the x values.
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
              do j = iStart, iFinis
                call pgmove(x(j), y(j, 1) - yErrors(j))
                call pgdraw(x(j), y(j, 1) + yErrors(j))
              end do

            else
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
            end if
            iStart = i - 1
            inNullPatch = .true.
          end if
        end do
        if (inNullPatch) then
          iFinis = vectorSize
          call pgsci(nullCI)
          call pgmove(x(iStart), 0.0)
          call pgdraw(x(iFinis), 0.0)
!print *, 'ci = nullCI:', nullCI
!print *, 'pgmove(x(',iStart,'), 0.0), pgdraw(x(',iFinis,'), 0.0)'

        else if (.not. inNullPatch) then
          iFinis = vectorSize
          call pgsci(plotCIs(1))
          if (withYerrors) then
!***** if too many x points, bin up the x values.
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
            do j = iStart, iFinis
              call pgmove(x(j), y(j, 1) - yErrors(j))
              call pgdraw(x(j), y(j, 1) + yErrors(j))
            end do

          else
!do j = iStart, iFinis
!  print *, j, x(j), y(j, 1)
!end do
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
          end if
        end if
      end if

    case('point', 'points')
!******* add square, diamond, circle??
      xEdgeWidth = 0.0
      yEdgeWidth = 0.05 * (maxY - minY)

      if (openNewDevice) then
        call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
        , minY - yEdgeWidth, maxY + yEdgeWidth)

        call pgsci(2)
        call pgmove(0.0, minY - yEdgeWidth)
        call pgdraw(0.0, maxY + yEdgeWidth)
        call pgmove(minX - xEdgeWidth, 0.0)
        call pgdraw(maxX + xEdgeWidth, 0.0)
        call pgsci(1)
        call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
        call pglab(trim(xLabel), trim(yLabel), trim(label))
      end if

      if (withYerrors) then
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgmove(x(i), y(i, 1) - yErrors(i))
            call pgdraw(x(i), y(i, 1) + yErrors(i))
          end if
        end do

      else
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgpt1(x(i), y(i, 1), -1)
          end if
        end do
      end if

    case('polygon')
      xEdgeWidth = 0.0
      yEdgeWidth = 0.05 * (maxY - minY)

      if (openNewDevice) then
        call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
        , minY - yEdgeWidth, maxY + yEdgeWidth)

        call pgsci(2)
        call pgmove(0.0, minY - yEdgeWidth)
        call pgdraw(0.0, maxY + yEdgeWidth)
        call pgmove(minX - xEdgeWidth, 0.0)
        call pgdraw(maxX + xEdgeWidth, 0.0)
        call pgsci(1)
        call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
        call pglab(trim(xLabel), trim(yLabel), trim(label))
      end if

      do plotNum = 1, numPlots
        call pgsci(plotCIs(plotNum))
        call pgpoly(vectorSize, x(1:vectorSize), y(1:vectorSize, plotNum))
      end do

    case default
      status = -5
  return
    end select

    if (wait .and. .not. openNewDevice) then
!print *, "pgslct at end of plotGraphKernel >>"
      call pgslct(previousDevId)
!print *, "<< pgslct at end of plotGraphKernel"
    else if (.not. wait) then
      call pgclos
    end if

  end function plotGraphKernel_old

  !---------------------------------------------------------------------
  function queryDeviceStatus(deviceId)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: deviceId
    integer :: queryDeviceStatus, answerLength, currentDeviceId, testDeviceId
    character(10) :: pgplotState

    !-------------------------------------------------------------------
    ! Start:

    if (deviceId .le. 0) then
      queryDeviceStatus = DEV_NOT_VALID ! not a valid device ID integer.

    else
      call pgqinf('STATE', pgplotState, answerLength)
      if (trim(pgplotState) .eq. 'OPEN') then
        call pgqid(currentDeviceId)
        if (deviceId .eq. currentDeviceId) then
          queryDeviceStatus = DEV_OPEN_AND_SEL ! deviceId is open and selected.

        else ! deviceId .ne. currentDeviceId
          call pgslct(deviceId)
          call pgqid(testDeviceId)
          if (testDeviceId .eq. currentDeviceId) then
            queryDeviceStatus = DEV_NOT_OPEN ! pgslct didn't succeed: ie, there
            ! are some pgplot devices open, but deviceId is not one of them.

          else
            call pgslct(currentDeviceId)
            queryDeviceStatus = DEV_OPEN_NOT_SEL ! deviceId is open but not
            ! currently selected.
          end if
        end if

      else
        queryDeviceStatus = NO_DEV_OPEN ! No pgplot devices are open.
      end if
    end if

  end function queryDeviceStatus

  !---------------------------------------------------------------------
  function queryDeviceStatus_old(deviceId)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: deviceId
    integer :: queryDeviceStatus_old, answerLength, currentDeviceId, testDeviceId
    character(10) :: pgplotState

    !-------------------------------------------------------------------
    ! Start:

    if (deviceId .le. 0) then
      queryDeviceStatus_old = -1 ! not a valid device ID integer. DEV_NOT_VALID

    else
      call pgqinf('STATE', pgplotState, answerLength)
      if (trim(pgplotState) .eq. 'OPEN') then
        call pgqid(currentDeviceId)
        if (deviceId .eq. currentDeviceId) then
          queryDeviceStatus_old = 3 ! deviceId is open and selected. DEV_OPEN_AND_SEL

        else ! deviceId .ne. currentDeviceId
!print *, "1st pgslct in queryDeviceStatus >>"
          call pgslct(deviceId)
!print *, "<< 1st pgslct in queryDeviceStatus"
          call pgqid(testDeviceId)
          if (testDeviceId .eq. currentDeviceId) then
            queryDeviceStatus_old = 1 ! pgslct didn't succeed, there are some
            ! pgplot devices open, but deviceId is not one of them. DEV_NOT_OPEN

          else
!print *, "2nd pgslct in queryDeviceStatus >>"
            call pgslct(currentDeviceId)
!print *, "<< 2nd pgslct in queryDeviceStatus"
            queryDeviceStatus_old = 2 ! deviceId is open but not currently selected. DEV_OPEN_NOT_SEL
          end if
        end if

      else
        queryDeviceStatus_old = 0 ! No pgplot devices are open. NO_DEV_OPEN
      end if
    end if

  end function queryDeviceStatus_old

  !---------------------------------------------------------------------
  subroutine findClrLimits(status, minStatus, maxStatus)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(out) :: status
    integer, intent(out), optional :: minStatus, maxStatus

    integer :: clrIndex
    real :: frac

    !-------------------------------------------------------------------
    ! Start

    status = 0

    minStatus = -1
    maxStatus = 0

    call pgqcol(minClrIndex, maxClrIndex)
    minClrIndex = max(minClrIndex, numReservedClrs)
    if (minClrIndex .ge. maxClrIndex) then
      status = -1
  return
    end if
    numClrIndices = maxClrIndex - minClrIndex + 1
    realNumCI = real(maxClrIndex + 1 - minClrIndex)
    realNumCIminus1 = real(maxClrIndex - minClrIndex)

    do clrIndex = numReservedClrs, 255
      frac = (clrIndex - numReservedClrs) / real(256 - numReservedClrs)
      oldXedges(clrIndex) = frac
    end do
    oldXedges(256) = 1.0

  end subroutine findClrLimits 






  !---------------------------------------------------------------------
  function angstroms2eV(angstroms) result(EeV)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: angstroms
    real(kind(1d0)) :: EeV

    !-------------------------------------------------------------------
    ! Start:

    EeV = PlanckH * SpeedOfLight / (angstroms * 1.0d-10) / ElectronQ

  end function angstroms2eV


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module temp_ssclib
